const fs = require('fs');
const Parser = require('junit2json').parse;
const crypto = require('crypto');
const Report = require('../lib/report.js')

class JUnit {
  constructor(options) {
    this.options = options;
    this.file = options.file;
    this.data = options.data;
    this.exclude = options.exclude;
    if (!this.exclude)
      this.exclude = [];
  }

  parseFrames(input) {
    /* Look at every line and guess what is a line number. */
    let lines = input.split("\n");
    let frames = [];
    let seen_dashes = 0;

    for (let i = 0; i < lines.length; i++) {
      let m = lines[i].match(/at ([A-za-z]+.*) \((.*\:[0-9]+.*)\)/);
      let function_name = null;
      let source = null;
      let source_file = null;

      if (lines[i].indexOf("_ _ _ _ _ _ _") > -1)
        seen_dashes++;

      if (!m || !m[1] || !m[2]) {
        if (seen_dashes !== 1)
          continue;

        /* Try an alternative pattern. */
        m = lines[i].match(/^(.*):([0-9]+):((in )?.*)$/);

        if (!m || !m[1] || !m[2])
          continue;

        if (m[3])
          function_name = m[3].substring(4);

        source = m[1];
        source_file = [m[1], m[2]];
      } else {
        function_name = m[1];
        source = m[2];
        source_file = m[2].split(":");
      }

      let frame = {};

      if (function_name)
        frame['function'] = function_name;

      if (source_file) {
        frame.file = source_file[0];
        frame.path = source_file[0];
      }

      if (source_file) {
        if (source_file.length >= 2)
          frame.line = parseInt(source_file[1]);
        if (source_file.length >= 3)
          frame.column = parseInt(source_file[2]);
      }

      frames.push(frame);
    }

    if (frames.length === 0) {
      frames.push({
        "function" : "Test case"
      });
    }

    return frames;
  }

  createReport(r, tc, ts, fingerprint) {
    let self = this;

    if (tc.failure && tc.failure.length > 0) {
      for (let i = 0; i < tc.failure.length; i++) {
        let report = new Report(r);
        let f = tc.failure[i];
        let signature_string = '';

        if (tc.classname) {
          report.set('attribute', 'unit.message', tc.classname);
          signature_string += tc.classname + '!';
        }

        if (ts.name) {
          signature_string += tc.name + '!';

          report.set('attribute', 'unit', ts.name);
          if (ts.file) {
            if (ts.file.endsWith(".js"))
              report.set('attribute', 'unit', ts.name + ".js");

            let category_token = ts.name.split("/");
            if (category_token.length > 1)
              report.set('attribute', 'category', category_token[0]);
          }
        }

        if (f.message) {
          signature_string += f.message + '!';
          if (f.message.indexOf('\n') >= 0)
            f.message = f.message.split('\n')[0];

          report.set('attribute', 'error.message', f.message);
        }

        if (f.type) {
          signature_string += f.type + '!';
          report.add('classifiers', f.type);
        } else if (f.message && f.message.indexOf(':') >= 0) {
          let cl_a = f.message.split(':');

          /* Stupid heuristic in case there is a junk pattern. */
          if (cl_a[0].length > 3)
            report.add('classifiers', cl_a[0]);
        }

        if (tc.time)
          report.set('attribute', 'duration', parseInt(tc.time * 1000));

        if (!fingerprint && signature_string.length > 0) {
          let rp = report.report;

          if (rp.attributes && rp.attributes['_mod_factor'])
            signature_string += rp.attributes[rp.attributes['_mod_factor']];

          fingerprint = crypto.createHash('sha256').update(signature_string).digest('hex');
          report.set('attribute', '_mod_fingerprint', fingerprint);
        } else if (fingerprint) {
          report.set('attribute', '_mod_fingerprint', fingerprint);
        }

        /* Synthesize a thread and a raw log buffer. */
        if (f.inner) {
          let thread = report.thread("junit");

          /*
           * parseFrames is guaranteed to return at least 1 frame, 
           * even if it's completely synthetic. We use this to attach
           * additional context. Finally, mela is responsible for submitting
           * a fingerprint value.
           */
          let frames = self.parseFrames(f.inner);
          for (let k = 0; k < frames.length; k++)
            thread.frame(frames[k]);

          let faulting_frame = frames[0];
          faulting_frame.text = f.inner;
        }
      }
    }
  }

  parseData(options, cb) {
    var self = this;
    var report = new Report();
    let fingerprint_set;

    report.checksum(self.data);

    report.add('classifiers', 'junit');
    report.set('attribute', 'tool', 'junit');

    if (options && options.attributes) {
      for (var a in options.attributes) {
        if (a === '_mod_fingerprint') {
          fingerprint_set = options,attributes[a];
        }

        report.set('attribute', a, options.attributes[a]);
      }
    }

    Parser(self.data).then(function(r) {
      if (!r.testsuite || r.testsuite.length === 0)
        return cb(new Error("no testsuite results found"), null);

      /* Extract failures. */
      for (let i = 0; i < r.testsuite.length; i++) {
        let ts_r = r.testsuite[i];

        if (ts_r.failures === 0)
          continue;

        for (let j = 0; j < ts_r.testcase.length; j++) {
          if (ts_r.testcase[j].failures === 0 || !ts_r.testcase[j].failure)
            continue;

          self.createReport(report, ts_r.testcase[j], ts_r, fingerprint_set);
        }
      }

      return cb(null, report);
    }).catch(function(e) {
      return cb(e, report);
    });
  }

  parse(options, cb) {
    var self = this;

    if (self.data) {
      return self.parseData(options, cb);
    } else if (self.file) {
      fs.readFile(this.file, "utf8", function(e, d) {
        if (e) {
          return cb(e, null);
        }

        self.data = d;
        return self.parseData(options, cb);
      });
    } else {
      return cb(new Error("no data was supplied"));
    }
  }
}

module.exports = JUnit;

//-- vim:ts=2:et:sw=2:st=2
