const fs = require('fs');
const { demangle } = require('demangle');

const Report = require('../lib/report.js')

const callstackHeaderRegex = / *# +Address +Function/;
const fullLineRegex = / [0-9]+ .*(0x[0-9A-F]+)[ ]{1,}(.*)!(.*) @ (.*)\(([0-9]+)\)/; 
const shortLineRegex = /[0-9]+ .*(0x[0-9A-F]+)[ ]{1,}(.*)!(.*)/;
const hexAddress = /0x[0-9A-F]+/;

class OrbisCoreView {
  constructor(options) {
    this.options = options;
    this.file = options.file;
    this.data = options.data;
    this.exclude = options.exclude;
    if (!this.exclude)
      this.exclude = [];
  }

  callstack(report, thread, lines, i) {
    var self = this;

    /*
     * If no thread is present in the debugger output, then just make
     * up a default name.
     */
    if (!thread) {
      thread = report.thread('Thread');
    }

    /* Extract every frame from the stack list command. */
    let is_callstack = false;
    for (; i < lines.length; i++) {
      if (lines[i].match(callstackHeaderRegex) !== null || 
         lines[i].indexOf('stack list') >= 0 ) {
        continue;
      }

      let line = lines[i].match(fullLineRegex);
      if (line == null) {
        // Try to parse for no path/filename/line number.
        line = lines[i].match(shortLineRegex);
        if (line == null) {
          continue;
        }
      }

      let frame;
      let file;
      
      // Extract file if we have it
      if (line.length > 4) {
        let file = line[4];
        let f_s = file.lastIndexOf("/");
        if (f_s === -1) {
          f_s = file.lastIndexOf("\\");
        }
        if (f_s >= 0) {
          file = file.substring(f_s + 1, file.length);
        }
      }

      let fn = line[3];
      if (fn[0] === '[')
        fn = fn.substring(fn.indexOf(']') + 2, fn.length); 
      
      // Repair mangled names
      if (fn.startsWith("_Z")) {
        let demangled = demangle("_" + fn);
        if (demangled) {
          fn = demangled;
        }
      }
      
      let address = fn.match(hexAddress);
      if (address !== null) {
        // Don't use an address as function name
        address = fn;
        fn = '';
      }
      
      frame = {
        registers: {
          ip: line[1]
        },
        'object' : line[2],
        'function': fn,
        'file': file,
        'path': line[4],
        'line': line[5],
        'address': address,
      };

      if (is_callstack === false) {
        frame.text = self.data + '\n';
        is_callstack = true;
      }
      thread.frame(frame);
    }

    return i;
  }

  parseData(options, cb) {
    let self = this;
    let data = self.data;

    /* Check for signature string first. */
    let is_coreview = data.indexOf('Core File Viewer for PlayStation');
    if (is_coreview === -1)
      is_coreview = data.indexOf('NEO Hardware: ');
    if (is_coreview === -1)
      return cb('invalid data detected');

    let report = new Report();
    report.checksum(data);

    if (options && options.attributes) {
      for (var a in options.attributes)
        report.set('attribute', a, options.attributes[a]);
    }

    if (data.indexOf("Core File Viewer for PlayStation®5") >= 0) {
      report.set('attribute', 'uname.sysname', 'ps5');
    } else {
      report.set('attribute', 'uname.sysname', 'ps4');
    }

    /* Dispatch to the appropriate handler depending on context. */
    let lines = data.split('\n');

    let path = null;
    let neo = null;
    let callstack = null;
    let thread = null;
    let classifiers = null;
    let classifiers_first = false;

    for (var i = 0; i < lines.length; i++) {
      if (path === null) {
        path = lines[i].match(/coreview .*> coredump load "(.*)"/);

        if (path !== null) {
          report.set('attribute', 'coredump', path[1]);
          continue;
        }
      }

      if (neo === null) {
        neo = lines[i].match(/NEO Hardware: (.*), NEO Mode: (.*)/);

        if (neo !== null) {
          report.set('attribute', 'neo.hardware', neo[1])
                .set('attribute', 'neo.mode', neo[2]);

          /* The subsequent line contains fault information. */
          continue;
        }
      }

      if (callstack === null) {
        callstack = lines[i].match(/coreview .*> stack list/);

        if (callstack !== null) {
          /* The subsequent lines contain the actual callstack. */
          i = self.callstack(report, thread, lines, i + 1);
          if (i === -1)
            return cb(new Error('callstack extraction failure'));

          continue;
        }
      }

      if (classifiers === null) {
        classifiers = lines[i].match(/^(0x[A-F0-9]+):(0x[A-F0-9]+): (0x[A-F0-9]+) /);
        classifiers_first = true;
      }

      if (classifiers_first === true && classifiers !== null) {
        classifiers_first = false;

        var message = lines[i].substring(classifiers[0].length, lines[i].length);

        message = message.replace('\r', '');
        message = message.replace('.', '');
        report.set('attribute', 'fault.address', parseInt(classifiers[3], 16))
              .set('attribute', 'error.message', message);

        let classifier = '';

        /*
         * These are common classifiers seen from Orbis.
         */
        if (message.indexOf('Bus error encountered') === 0) {
          classifier = 'bus memory';
        } else if (message.indexOf('abort() has been') === 0) {
          classifier = 'abort';
        } else if (message.indexOf('GPU encountered illegal') === 0) {
          classifier = 'illegal-instruction gpu';
        } else if (message.indexOf('GPU page fault') === 0) {
          classifier = 'page-fault gpu';
        } else if (message.indexOf('Segmentation violation') === 0) {
          classifier = 'segv memory';
        } else if (message.indexOf('user stack is corrupted') >= 0) {
          classifier = 'stack-corruption memory';
        } else if (message.indexOf('forcibly shutdown the process') >= 0) {
          classifier = 'watchdog';
        } else if (message.indexOf('The process was terminated with') >= 0) {
          classifier = 'abnormal-termination';
        } else {
          /*
           * Take first three words in message and treat as classifier. This is
           * pending one of PS4 documentation or CoreView string dump.
           */
          var c_s = message.split(' ');

          for (let j = 0; j < 3; j++) {
            classifier += c_s[j];
            if (j !== 2)
              classifier += '-';
          }

          if (message.indexOf('GPU') === 0) {
            classifier += ' gpu';
          }
        }

        report.add("classifiers", classifier.toLowerCase());
      }
    }

    report.reports = [report];
    return cb(null, report);
  }

  parse(options, cb) {
    var self = this;

    if (self.data) {
      return self.parseData(options, cb);
    } else if (self.file) {
      fs.readFile(this.file, "utf8", function(e, d) {
        if (e) {
          return cb(e, null);
        }

        self.data = d;
        return self.parseData(options, cb);
      });
    } else {
      return cb(new Error("no data was supplied"));
    }
  }
}

module.exports = OrbisCoreView;

//-- vim:ts=2:et:sw=2:st=2
