#!/usr/bin/env node

const Valgrind = require('../lib/valgrind.js');
const Sanitizers = require('../lib/sanitizers.js');
const DrMemory = require('../lib/drmemory.js');
const JUnit = require('../lib/junit.js');
const OrbisCoreView = require('../lib/orbis.js');
const Pool = require('../lib/pool.js');
const bodyParser = require('body-parser');
const Express = require('express');
const cors = require('cors');
const cluster = require('cluster');
const axios = require('axios');
const https = require('https');
const http = require('http');
const fs = require('fs');
const { BacktraceClient } = require('@backtrace/node');

const paths = [
  '/etc/backtrace/mela/mela.conf',
  'assets/mela.conf',
  '../assets/mela.conf',
];
let configuration_path;
let configuration;

const pools = {};

var metrics = {};

function metrics_initialize() {
  var formats = [
    'junit',
    'drmemory',
    'valgrind',
    'sanitizers',
    'orbiscoreview',
  ];
  metrics.reports = {};
  metrics.forward = {};
  metrics.target_verbose = {};
  metrics.target_count = { _type: 'L', _value: 0 };

  for (var i = 0; i < formats.length; i++) {
    metrics.reports[formats[i]] = {
      _type: 'L',
      _value: 0,
    };

    metrics.forward[formats[i]] = {
      _type: 'L',
      _value: 0,
    };
  }

  metrics.forward.queue = { _type: 'L', _value: 0 };
  metrics.forward.success = { _type: 'L', _value: 0 };
  metrics.forward.failed = { _type: 'L', _value: 0 };
  return;
}

function send(request, response, reports, object, universe, token, format) {
  let url;
  let count = reports.length;
  var concurrency = 10;

  if (configuration.forward.concurrency)
    concurrency = parseInt(configuration.forward.concurrency);

  url = configuration.forward.target.replace('{universe}', universe);
  url = url.replace('{token}', token);

  var hostname = url.split('/')[2];
  pools[hostname] = pools[hostname] || new Pool(concurrency);

  if (!metrics.target_verbose[hostname]) {
    metrics.target_verbose[hostname] = { _type: 'L', _value: 0 };
    metrics.target_count._value++;
  }

  if (reports.length > 0 && reports[0].attributes) {
    if (reports[0].attributes['_mod_sync'])
      url += '&_mod_sync=' + reports[0].attributes['_mod_sync'];
    else if (reports[0].attributes['sync'])
      url += '&sync=' + reports[0].attributes['sync'];
  }

  metrics.target_verbose[hostname]._value += 1;

  object.failures = 0;
  object.created = 0;
  object.kind = {};
  object.resource = url;
  object.responses = [];

  /* Construct the URL from the configuration template. */
  for (let i = 0; i < reports.length; i++) {
    for (let j = 0; j < reports[i].classifiers.length; j++) {
      if (!object.kind[reports[i].classifiers[j]])
        object.kind[reports[i].classifiers[j]] = 0;
      object.kind[reports[i].classifiers[j]]++;
    }

    if (request.query && request.query['timestamp']) {
      reports[i].timestamp = parseInt(request.query['timestamp']);
    }

    pools[hostname].add(function () {
      return new Promise(function (resolve) {
        axios
          .post(url, reports[i])
          .then((r) => {
            object.responses.push(r.data);
            object.created++;
            count--;
            metrics.forward.queue._value--;
            metrics.forward.success._value++;

            if (count === 0) response.send(object);

            resolve();
          })
          .catch((error) => {
            count--;
            object.failures++;
            metrics.forward.queue._value--;
            metrics.forward.failed._value++;

            if (count === 0) {
              response.send(object);
            }

            resolve();
          });
      });
    });
  }

  return;
}

function start(argv) {
  for (let i = 0; i < paths.length; i++) {
    if (fs.existsSync(paths[i])) {
      configuration_path = paths[i];
      break;
    }
  }

  if (!configuration_path) {
    console.error('No configuration file found');
    return;
  }

  if (cluster.isMaster) {
    console.log('Master cluster is starting up worker...');
    cluster.fork();
    cluster.on('exit', (worker, code, signal) => {
      console.error('Worker has died, restarting it.');
      cluster.fork();
    });

    return;
  }

  process.on('uncaughtException', function (error) {
    console.error(error + '');
  });

  if (!cluster.isWorker) return;

  console.log('Configuration file found: ' + configuration_path);
  configuration = JSON.parse(fs.readFileSync(configuration_path));

  let backtraceClient;
  if (configuration.backtrace) {
    backtraceClient = BacktraceClient.initialize({
      url: configuration.backtrace.endpoint,
      metrics: {
        ...(configuration.backtrace?.metrics ?? {}),
        metricsSubmissionUrl: 'https://events-test.backtrace.io',
      },
    });
  }

  const express = Express();
  express.use(
    bodyParser.raw({
      limit: '100mb',
      type: '*/*',
    }),
  );
  express.use(cors());

  express.use(function (error, request, response, next) {
    if (response.headersSent) {
      return next(error);
    }
    if (backtraceClient) {
      backtraceClient.send(error, {
        originalUrl: request.originalUrl,
        protocol: request.protocol,
        query: request.query,
        headers: request.headers,
        body: { value: request.body }, // Avoid putting string bodies in attributes
      });
    }

    response.status(500).send({ error: 'malformed input' });
    return;
  });

  metrics_initialize();
  if (configuration.circonus) {
    if (!configuration.circonus.uuid || !configuration.circonus.secret) {
      console.error('Circonus uuid and secret are required.');
    }
  }

  const port =
    process.env.NODE_ENV == 'production' ? configuration.bind.port : '2345';
  if (
    process.env.NODE_ENV == 'production' &&
    configuration.bind &&
    configuration.bind.ssl &&
    configuration.bind.ssl.key &&
    configuration.bind.ssl.key !== 'path'
  ) {
    const options = {
      key: fs.readFileSync(configuration.bind.ssl.key),
      cert: fs.readFileSync(configuration.bind.ssl.certificate),
    };

    server = https.createServer(options, express).listen(port, function () {
      console.log('[HTTPS] Service started at port ' + port);
    });
  } else {
    server = http.createServer(express).listen(port, function () {
      console.log('[HTTP] Listening on port ' + port);
    });
  }

  express.get('/api/metrics', function (request, response) {
    if (
      !request.query ||
      !request.query.token ||
      !configuration.circonus ||
      !configuration.circonus.secret ||
      configuration.circonus.secret != request.query.token
    ) {
      return response.status(500).send({ error: 'invalid token' });
    }

    return response.send(metrics);
  });

  express.get('/api/mela/status', function (request, response) {
    response.send({ status: 'ok' });
  });

  express.post(configuration.forward.route, (request, response) => {
    const start = Date.now();

    /* The first step is to validate the universe, token and format. */
    const format = request.params.format;
    const universe = request.params.universe;
    const token = request.params.token;

    let attributes = {};
    for (let k in request.query) {
      attributes[k] = request.query[k];
    }

    let parser;

    if (format === 'valgrind') {
      parser = new Valgrind({ data: request.body.toString('utf8') });
    } else if (format === 'junit') {
      parser = new JUnit({ data: request.body.toString('utf8') });
    } else if (format === 'sanitizers') {
      parser = new Sanitizers({ data: request.body.toString('utf8') });
    } else if (format === 'drmemory') {
      parser = new DrMemory({ data: request.body.toString('utf8') });
    } else if (format === 'orbiscoreview') {
      parser = new OrbisCoreView({ data: request.body.toString('utf8') });
    } else {
      response.status(500).send({ error: 'unknown format' });
      return;
    }

    metrics.reports[format]._value++;

    parser.parse({ attributes: attributes }, (error, reports) => {
      const st = { status: 'ok' };
      const latency = {};
      let now;

      if (error) {
        if (backtraceClient) {
          backtraceClient.send(error, {
            originalUrl: request.originalUrl,
            protocol: request.protocol,
            query: request.query,
            headers: request.headers,
            body: { value: request.body },
          });
        }
        return response.status(500).send({ error: error + '' });
      }

      now = latency.now = Date.now();
      latency.parse = now - start;

      /* Now, materialize the results. */
      const result = reports.serialize();

      now = Date.now();
      latency.serialize = now - latency.now;
      latency.now = now;

      st.count = result.length;
      metrics.forward.queue._value += result.length;

      /* Extract the UUID of the report. */
      if (result.length === 0) {
        return response.send(st);
      }

      st.guid = result[0].uuid;

      /* Forward the actual content to the server. */
      delete latency.now;
      st.latency = latency;

      return send(request, response, result, st, universe, token, format);
    });

    return;
  });

  return;
}

module.exports = start;

//-- vim:ts=2:et:sw=2
