"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BacktraceCoreClient = void 0;
const _1 = require(".");
const Events_1 = require("./common/Events");
const BacktraceReportSubmission_1 = require("./model/http/BacktraceReportSubmission");
const BacktraceReport_1 = require("./model/report/BacktraceReport");
const AttributeManager_1 = require("./modules/attribute/AttributeManager");
const ClientAttributeProvider_1 = require("./modules/attribute/ClientAttributeProvider");
const UserAttributeProvider_1 = require("./modules/attribute/UserAttributeProvider");
const BreadcrumbsManager_1 = require("./modules/breadcrumbs/BreadcrumbsManager");
const V8StackTraceConverter_1 = require("./modules/converter/V8StackTraceConverter");
const BacktraceDataBuilder_1 = require("./modules/data/BacktraceDataBuilder");
const BacktraceDatabase_1 = require("./modules/database/BacktraceDatabase");
const BacktraceDatabaseFileStorageProvider_1 = require("./modules/database/BacktraceDatabaseFileStorageProvider");
const BacktraceMetrics_1 = require("./modules/metrics/BacktraceMetrics");
const MetricsBuilder_1 = require("./modules/metrics/MetricsBuilder");
const SingleSessionProvider_1 = require("./modules/metrics/SingleSessionProvider");
const RateLimitWatcher_1 = require("./modules/rateLimiter/RateLimitWatcher");
class BacktraceCoreClient {
    /**
     * Determines if the client is enabled.
     */
    get enabled() {
        return this._enabled;
    }
    /**
     * Current session id
     */
    get sessionId() {
        return this._sessionProvider.sessionId;
    }
    /**
     * Backtrace SDK name
     */
    get agent() {
        return this._sdkOptions.agent;
    }
    /**
     * Backtrace SDK version
     */
    get agentVersion() {
        return this._sdkOptions.agentVersion;
    }
    /**
     * Available cached client attributes
     */
    get attributes() {
        return this.attributeManager.get().attributes;
    }
    /**
     * Available cached client annotatations
     */
    get annotations() {
        return this.attributeManager.get().annotations;
    }
    get metrics() {
        return this._modules.get(BacktraceMetrics_1.BacktraceMetrics);
    }
    get breadcrumbs() {
        return this._modules.get(BreadcrumbsManager_1.BreadcrumbsManager);
    }
    get database() {
        return this._modules.get(BacktraceDatabase_1.BacktraceDatabase);
    }
    /**
     * Client cached attachments
     */
    get attachments() {
        // always return a copy of attachments
        return [...this._attachments];
    }
    /**
     * Modules used by the client
     */
    get modules() {
        return this._modules;
    }
    get sessionFiles() {
        return this._modules.get(_1.SessionFiles);
    }
    constructor(setup) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j;
        this._modules = new Map();
        this._enabled = false;
        this.reportEvents = new Events_1.Events();
        this.options = setup.options;
        this.fileSystem = setup.fileSystem;
        this._sdkOptions = setup.sdkOptions;
        this._attachments = (_a = this.options.attachments) !== null && _a !== void 0 ? _a : [];
        this._sessionProvider = (_b = setup.sessionProvider) !== null && _b !== void 0 ? _b : new SingleSessionProvider_1.SingleSessionProvider();
        this._reportSubmission =
            (_c = setup.reportSubmission) !== null && _c !== void 0 ? _c : new BacktraceReportSubmission_1.RequestBacktraceReportSubmission(this.options, setup.requestHandler);
        this._rateLimitWatcher = new RateLimitWatcher_1.RateLimitWatcher(this.options.rateLimit);
        this._requestHandler = setup.requestHandler;
        const attributeProviders = [
            new ClientAttributeProvider_1.ClientAttributeProvider(this.agent, this.agentVersion, this._sessionProvider.sessionId),
        ];
        if (setup.attributeProviders) {
            attributeProviders.push(...setup.attributeProviders);
        }
        if (this.options.userAttributes) {
            attributeProviders.push(new UserAttributeProvider_1.UserAttributeProvider(this.options.userAttributes));
        }
        this.attributeManager = new AttributeManager_1.AttributeManager(attributeProviders);
        const stackTraceConverter = (_d = setup.stackTraceConverter) !== null && _d !== void 0 ? _d : new V8StackTraceConverter_1.V8StackTraceConverter();
        this._dataBuilder = new BacktraceDataBuilder_1.BacktraceDataBuilder(this._sdkOptions, stackTraceConverter, this.attributeManager, new _1.DebugIdProvider(stackTraceConverter, setup.debugIdMapProvider));
        if (((_f = (_e = this.options) === null || _e === void 0 ? void 0 : _e.database) === null || _f === void 0 ? void 0 : _f.enable) === true && setup.fileSystem) {
            const provider = BacktraceDatabaseFileStorageProvider_1.BacktraceDatabaseFileStorageProvider.createIfValid(setup.fileSystem, this.options.database);
            if (this.fileSystem) {
                const sessionFiles = new _1.SessionFiles(this.fileSystem, this.options.database.path, this.sessionId, (_g = this.options.database.maximumOldSessions) !== null && _g !== void 0 ? _g : 1);
                this._modules.set(_1.SessionFiles, sessionFiles);
            }
            if (provider) {
                const database = new BacktraceDatabase_1.BacktraceDatabase(this.options.database, provider, this._reportSubmission, this.sessionFiles);
                this._modules.set(BacktraceDatabase_1.BacktraceDatabase, database);
            }
        }
        const metrics = new MetricsBuilder_1.MetricsBuilder(this.options, this._sessionProvider, this.attributeManager, setup.requestHandler).build(setup.uniqueMetricsQueue, setup.summedMetricsQueue);
        if (metrics) {
            this._modules.set(BacktraceMetrics_1.BacktraceMetrics, metrics);
        }
        if (((_h = this.options.breadcrumbs) === null || _h === void 0 ? void 0 : _h.enable) !== false) {
            const breadcrumbsManager = new BreadcrumbsManager_1.BreadcrumbsManager((_j = this.options) === null || _j === void 0 ? void 0 : _j.breadcrumbs, setup.breadcrumbsSetup);
            this._modules.set(BreadcrumbsManager_1.BreadcrumbsManager, breadcrumbsManager);
        }
        if (setup.modules) {
            for (const module of setup.modules) {
                this.addModule(module);
            }
        }
    }
    initialize() {
        var _a, _b, _c;
        if (this.fileSystem && ((_a = this.options.database) === null || _a === void 0 ? void 0 : _a.createDatabaseDirectory)) {
            if (!this.options.database.path) {
                throw new Error('Missing mandatory path to the database. Please define the database.path option in the configuration.');
            }
            this.fileSystem.createDirSync((_b = this.options.database) === null || _b === void 0 ? void 0 : _b.path);
        }
        for (const module of this._modules.values()) {
            if (module.bind) {
                module.bind(this.getModuleBindData());
            }
            if (module.initialize) {
                module.initialize();
            }
        }
        (_c = this.sessionFiles) === null || _c === void 0 ? void 0 : _c.clearPreviousSessions();
        this._enabled = true;
    }
    addAttribute(attributes) {
        this.attributeManager.add(attributes);
    }
    /**
     * Add attachment to the client
     * @param attachment attachment
     */
    addAttachment(attachment) {
        this._attachments.push(attachment);
    }
    // This function CANNOT be an async function due to possible async state machine stack frame inclusion, which breaks the skip stacks
    send(data, reportAttributesOrAbortSignal, reportAttachments = [], abortSignal) {
        if (!this._enabled) {
            return Promise.resolve();
        }
        if (this._rateLimitWatcher.skipReport()) {
            return Promise.resolve();
        }
        // If data is BacktraceReport, we know that the second argument should be only AbortSignal
        const reportAttributes = !this.isReport(data)
            ? reportAttributesOrAbortSignal
            : undefined;
        // If data is BacktraceReport, we know that the second argument should be only AbortSignal
        abortSignal = !this.isReport(data) ? abortSignal : reportAttributesOrAbortSignal;
        const report = this.isReport(data)
            ? data
            : new BacktraceReport_1.BacktraceReport(data, reportAttributes, [], {
                skipFrames: this.skipFrameOnMessage(data),
            });
        this.reportEvents.emit('before-skip', report);
        if (this.options.skipReport && this.options.skipReport(report)) {
            return Promise.resolve();
        }
        const backtraceData = this.generateSubmissionData(report);
        if (!backtraceData) {
            return Promise.resolve();
        }
        const submissionAttachments = this.generateSubmissionAttachments(report, reportAttachments);
        this.reportEvents.emit('before-send', report, backtraceData, submissionAttachments);
        return this._reportSubmission
            .send(backtraceData, submissionAttachments, abortSignal)
            .then((submissionResult) => {
            this.reportEvents.emit('after-send', report, backtraceData, submissionAttachments, submissionResult);
        });
    }
    /**
     * Disposes the client and all client callbacks
     */
    dispose() {
        this._enabled = false;
        for (const module of this._modules.values()) {
            if (module.dispose) {
                module.dispose();
            }
        }
    }
    addModule(typeOrModule, module) {
        let type;
        if (typeof typeOrModule === 'function') {
            type = typeOrModule;
        }
        else {
            module = typeOrModule;
            type = Object.getPrototypeOf(module);
        }
        if (!module) {
            throw new Error('Module implementation is required.');
        }
        this._modules.set(type, module);
        if (this._enabled) {
            module.bind && module.bind(this.getModuleBindData());
            module.initialize && module.initialize();
        }
    }
    generateSubmissionData(report) {
        const backtraceData = this._dataBuilder.build(report);
        if (!this.options.beforeSend) {
            return backtraceData;
        }
        return this.options.beforeSend(backtraceData);
    }
    generateSubmissionAttachments(report, reportAttachments) {
        var _a;
        return [...this._attachments, ...((_a = report.attachments) !== null && _a !== void 0 ? _a : []), ...(reportAttachments !== null && reportAttachments !== void 0 ? reportAttachments : [])];
    }
    skipFrameOnMessage(data) {
        return typeof data === 'string' ? 1 : 0;
    }
    isReport(data) {
        return data instanceof BacktraceReport_1.BacktraceReport;
    }
    getModuleBindData() {
        return {
            client: this,
            options: this.options,
            reportEvents: this.reportEvents,
            attributeManager: this.attributeManager,
            reportSubmission: this._reportSubmission,
            requestHandler: this._requestHandler,
            sessionFiles: this.sessionFiles,
        };
    }
}
exports.BacktraceCoreClient = BacktraceCoreClient;
//# sourceMappingURL=BacktraceCoreClient.js.map