"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BacktraceDatabase = void 0;
const IdGenerator_1 = require("../../common/IdGenerator");
const TimeHelper_1 = require("../../common/TimeHelper");
const BacktraceDatabaseContext_1 = require("./BacktraceDatabaseContext");
class BacktraceDatabase {
    /**
     * Determines if the database is enabled.
     */
    get enabled() {
        return this._enabled;
    }
    constructor(_options, _storageProvider, _requestHandler, _sessionFiles) {
        var _a, _b, _c, _d, _e;
        this._options = _options;
        this._storageProvider = _storageProvider;
        this._requestHandler = _requestHandler;
        this._sessionFiles = _sessionFiles;
        this._storageProviders = [];
        this._enabled = false;
        this._databaseRecordContext = new BacktraceDatabaseContext_1.BacktraceDatabaseContext((_a = this._options) === null || _a === void 0 ? void 0 : _a.maximumRetries);
        this._maximumRecords = (_c = (_b = this._options) === null || _b === void 0 ? void 0 : _b.maximumNumberOfRecords) !== null && _c !== void 0 ? _c : 8;
        this._retryInterval = (_e = (_d = this._options) === null || _d === void 0 ? void 0 : _d.retryInterval) !== null && _e !== void 0 ? _e : 60000;
    }
    /**
     * Starts database integration.
     * @returns true if the database started successfully. Otherwise false.
     */
    initialize() {
        var _a, _b;
        if (this._enabled) {
            return this._enabled;
        }
        if (((_a = this._options) === null || _a === void 0 ? void 0 : _a.enable) === false) {
            return false;
        }
        const startResult = this._storageProvider.start();
        if (!startResult) {
            return false;
        }
        const lockId = (_b = this._sessionFiles) === null || _b === void 0 ? void 0 : _b.lockPreviousSessions();
        this.loadReports()
            .then(() => {
            this.setupDatabaseAutoSend();
        })
            .finally(() => { var _a; return lockId && ((_a = this._sessionFiles) === null || _a === void 0 ? void 0 : _a.unlockPreviousSessions(lockId)); });
        this._enabled = true;
        return true;
    }
    bind({ reportEvents }) {
        var _a;
        if (this._enabled) {
            return;
        }
        if (((_a = this._options) === null || _a === void 0 ? void 0 : _a.enable) === false) {
            return;
        }
        reportEvents.on('before-send', (_, data, attachments) => {
            const record = this.add(data, attachments);
            if (!record || record.locked || record.count !== 1) {
                return undefined;
            }
            record.locked = true;
        });
        reportEvents.on('after-send', (_, data, __, submissionResult) => {
            var _a;
            const record = this._databaseRecordContext.find((record) => record.data.uuid === data.uuid);
            if (!record) {
                return;
            }
            record.locked = false;
            if (submissionResult.status === 'Ok') {
                this.remove(record);
                (_a = this._sessionFiles) === null || _a === void 0 ? void 0 : _a.unlockPreviousSessions(record.id);
            }
        });
    }
    /**
     * Adds backtrace data to the database
     * @param backtraceData diagnostic data object
     * @param attachments attachments
     * @returns record if database is enabled. Otherwise undefined.
     */
    add(backtraceData, attachments) {
        if (!this._enabled) {
            return undefined;
        }
        this.prepareDatabase();
        const record = {
            count: 1,
            data: backtraceData,
            timestamp: TimeHelper_1.TimeHelper.now(),
            hash: '',
            id: IdGenerator_1.IdGenerator.uuid(),
            locked: false,
            attachments: attachments,
        };
        const saveResult = this._storageProvider.add(record);
        if (!saveResult) {
            return undefined;
        }
        this._databaseRecordContext.add(record);
        this.lockSessionWithRecord(record);
        return record;
    }
    /**
     * Returns stored references to Backtrace records
     * @returns available records in the database
     */
    get() {
        return this._databaseRecordContext.get();
    }
    /**
     * @returns Returns number of records stored in the Database
     */
    count() {
        return this._databaseRecordContext.count();
    }
    /**
     * Disables database integration. After running the dispose method, you cannot
     * execute any database operations.
     */
    dispose() {
        this._enabled = false;
        clearInterval(this._intervalId);
    }
    /**
     * Removes the database record/records
     * @param recordOrRecords database records
     */
    remove(recordOrRecords) {
        var _a;
        if (!this._enabled) {
            return;
        }
        const records = Array.isArray(recordOrRecords) ? recordOrRecords : [recordOrRecords];
        for (const record of records) {
            this._databaseRecordContext.remove(record);
            this._storageProvider.delete(record);
            (_a = this._sessionFiles) === null || _a === void 0 ? void 0 : _a.unlockPreviousSessions(record.id);
        }
    }
    addStorageProvider(storageProvider) {
        if (this._enabled) {
            throw new Error('Cannot add storage provider after the database has been enabled.');
        }
        this._storageProviders.push(storageProvider);
    }
    /**
     * Sends all records available in the database to Backtrace and removes them
     * no matter if the submission process was successful or not.
     */
    flush() {
        return __awaiter(this, void 0, void 0, function* () {
            const start = TimeHelper_1.TimeHelper.now();
            yield this.send();
            const records = this.get().filter((n) => n.timestamp <= start);
            for (const record of records) {
                this.remove(record);
            }
        });
    }
    /**
     * Sends all records available in the database to Backtrace.
     */
    send() {
        return __awaiter(this, void 0, void 0, function* () {
            for (let bucketIndex = 0; bucketIndex < this._databaseRecordContext.bucketCount; bucketIndex++) {
                // make a copy of records to not update the array after each remove
                const records = [...this._databaseRecordContext.getBucket(bucketIndex)];
                for (const record of records) {
                    if (record.locked) {
                        continue;
                    }
                    try {
                        record.locked = true;
                        const result = yield this._requestHandler.send(record.data, record.attachments);
                        if (result.status === 'Ok') {
                            this.remove(record);
                            continue;
                        }
                        this._databaseRecordContext.increaseBucket(bucketIndex);
                        return;
                    }
                    finally {
                        record.locked = false;
                    }
                }
            }
        });
    }
    /**
     * Prepare database to insert records
     * @param totalNumberOfRecords number of records to insert
     */
    prepareDatabase(totalNumberOfRecords = 1) {
        const numberOfRecords = this.count();
        if (numberOfRecords + totalNumberOfRecords <= this._maximumRecords) {
            return;
        }
        this.remove(this._databaseRecordContext.dropOverflow(totalNumberOfRecords));
    }
    loadReports() {
        return __awaiter(this, void 0, void 0, function* () {
            const records = yield this._storageProvider.get();
            // delete old records before adding them to the database
            if (records.length >= this._maximumRecords) {
                this.remove(records.splice(this._maximumRecords));
            }
            this.prepareDatabase(records.length);
            this._databaseRecordContext.load(records);
            for (const record of records) {
                this.lockSessionWithRecord(record);
            }
        });
    }
    setupDatabaseAutoSend() {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            if (((_a = this._options) === null || _a === void 0 ? void 0 : _a.autoSend) === false) {
                return;
            }
            const sendDatabaseReports = () => __awaiter(this, void 0, void 0, function* () {
                yield this.send();
            });
            this._intervalId = setInterval(sendDatabaseReports, this._retryInterval);
            yield this.send();
        });
    }
    lockSessionWithRecord(record) {
        var _a;
        if (!this._sessionFiles) {
            return;
        }
        const sessionId = (_a = record.data.attributes) === null || _a === void 0 ? void 0 : _a['application.session'];
        if (typeof sessionId !== 'string') {
            this._sessionFiles.lockPreviousSessions(record.id);
            return;
        }
        const session = this._sessionFiles.getSessionWithId(sessionId);
        session === null || session === void 0 ? void 0 : session.lock(record.id);
    }
}
exports.BacktraceDatabase = BacktraceDatabase;
//# sourceMappingURL=BacktraceDatabase.js.map