"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BacktraceMetrics = void 0;
const AbortController_1 = require("../../common/AbortController");
const AbortError_1 = require("../../common/AbortError");
const TimeHelper_1 = require("../../common/TimeHelper");
const ReportDataBuilder_1 = require("../attribute/ReportDataBuilder");
const SummedEvent_1 = require("./model/SummedEvent");
const UniqueEvent_1 = require("./model/UniqueEvent");
class BacktraceMetrics {
    /**
     * Returns current session id.
     */
    get sessionId() {
        return this._sessionProvider.sessionId;
    }
    constructor(_options, _sessionProvider, _attributeManager, _summedEventsSubmissionQueue, _uniqueEventsSubmissionQueue) {
        var _a, _b;
        this._options = _options;
        this._sessionProvider = _sessionProvider;
        this._attributeManager = _attributeManager;
        this._summedEventsSubmissionQueue = _summedEventsSubmissionQueue;
        this._uniqueEventsSubmissionQueue = _uniqueEventsSubmissionQueue;
        /**
         * Default metrics submission interval. The variable defines how often metrics will be sent to metrics system.
         */
        this.DEFAULT_UPDATE_INTERVAL = TimeHelper_1.TimeHelper.convertSecondsToMilliseconds(30 * 60);
        this.DEFAULT_SERVER_URL = 'https://events.backtrace.io';
        this.metricsHost = (_a = this._options.metricsSubmissionUrl) !== null && _a !== void 0 ? _a : this.DEFAULT_SERVER_URL;
        this._updateInterval = (_b = this._options.autoSendInterval) !== null && _b !== void 0 ? _b : this.DEFAULT_UPDATE_INTERVAL;
        this._abortController = (0, AbortController_1.createAbortController)();
    }
    /**
     * Starts metrics submission.
     */
    initialize() {
        if (!this._sessionProvider.newSession) {
            return;
        }
        this.addSummedEvent('Application Launches');
        this.handleAbort(() => this.send(this._abortController.signal));
        if (this._updateInterval === 0) {
            return;
        }
        this._updateIntervalId = setInterval(() => {
            this.handleAbort(() => this.send(this._abortController.signal));
        }, this._updateInterval);
    }
    /**
     * Returns total number of events in the submission queue.
     */
    count() {
        var _a, _b, _c, _d;
        return (_d = (_b = (_a = this._summedEventsSubmissionQueue) === null || _a === void 0 ? void 0 : _a.total) !== null && _b !== void 0 ? _b : 0 + ((_c = this._uniqueEventsSubmissionQueue) === null || _c === void 0 ? void 0 : _c.total)) !== null && _d !== void 0 ? _d : 0;
    }
    /**
     * Add summed event to next Backtrace Metrics request.
     * @param metricName Summed event name.
     * @param eventAttributes event attributes.
     */
    addSummedEvent(metricName, eventAttributes = {}) {
        if (!metricName) {
            return false;
        }
        const attributes = this.convertAttributes(Object.assign(Object.assign({}, this._attributeManager.get().attributes), ReportDataBuilder_1.ReportDataBuilder.build(eventAttributes !== null && eventAttributes !== void 0 ? eventAttributes : {}).attributes));
        this._summedEventsSubmissionQueue.add(new SummedEvent_1.SummedEvent(metricName, attributes));
        return true;
    }
    /**
     * Sends event to the metrics system.
     */
    send(abortSignal) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this._sessionProvider.shouldSend()) {
                return false;
            }
            yield Promise.all([this.sendUniqueEvent(abortSignal), this._summedEventsSubmissionQueue.send(abortSignal)]);
            this._sessionProvider.afterMetricsSubmission();
            return true;
        });
    }
    /**
     * Cleans up metrics interface.
     */
    dispose() {
        this._abortController.abort();
        if (this._updateIntervalId) {
            clearInterval(this._updateIntervalId);
        }
        this._uniqueEventsSubmissionQueue.dispose && this._uniqueEventsSubmissionQueue.dispose();
        this._summedEventsSubmissionQueue.dispose && this._summedEventsSubmissionQueue.dispose();
    }
    sendUniqueEvent(abortSignal) {
        return __awaiter(this, void 0, void 0, function* () {
            // always add the same unique event before send.
            const { attributes } = this._attributeManager.get();
            this._uniqueEventsSubmissionQueue.add(new UniqueEvent_1.UniqueEvent(this.convertAttributes(attributes)));
            yield this._uniqueEventsSubmissionQueue.send(abortSignal);
        });
    }
    /**
     * Event aggregators expecting to retrieve attributes in a string format. They also
     * don't expect to retrieve null/undefined as attribute values.
     */
    convertAttributes(attributes) {
        return Object.keys(attributes)
            .filter((n) => attributes[n] != null)
            .reduce((acc, n) => {
            var _a;
            acc[n] = (_a = attributes[n]) === null || _a === void 0 ? void 0 : _a.toString();
            return acc;
        }, {});
    }
    handleAbort(fn) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield fn();
                return true;
            }
            catch (err) {
                if (err instanceof AbortError_1.AbortError) {
                    return false;
                }
                throw err;
            }
        });
    }
}
exports.BacktraceMetrics = BacktraceMetrics;
//# sourceMappingURL=BacktraceMetrics.js.map