"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MetricsBuilder = void 0;
const TimeHelper_1 = require("../../common/TimeHelper");
const BacktraceMetrics_1 = require("./BacktraceMetrics");
const MetricsSubmissionQueue_1 = require("./MetricsSubmissionQueue");
const MetricsUrlInformation_1 = require("./MetricsUrlInformation");
class MetricsBuilder {
    constructor(_options, _sessionProvider, _attributeManager, _requestHandler) {
        this._options = _options;
        this._sessionProvider = _sessionProvider;
        this._attributeManager = _attributeManager;
        this._requestHandler = _requestHandler;
        /**
         * Default metrics submission interval. The variable defines how often metrics will be sent to metrics system
         * By default 30 mins.
         */
        this.DEFAULT_UPDATE_INTERVAL = TimeHelper_1.TimeHelper.convertSecondsToMilliseconds(30 * 60);
        this.APPLICATION_VERSION_ATTRIBUTE = 'application.version';
        this.APPLICATION_ATTRIBUTE = 'application';
    }
    build(uniqueEventsSubmissionQueue, summedEventsSubmissionQueue) {
        var _a;
        const metricsOptions = Object.assign(Object.assign({}, this.optionsWithDefaults()), ((_a = this._options.metrics) !== null && _a !== void 0 ? _a : {}));
        if (!metricsOptions.enable) {
            return undefined;
        }
        const applicationInfo = this.verifyAttributeSetup();
        if (!applicationInfo) {
            return undefined;
        }
        uniqueEventsSubmissionQueue =
            uniqueEventsSubmissionQueue !== null && uniqueEventsSubmissionQueue !== void 0 ? uniqueEventsSubmissionQueue : this.createUniqueEventSubmissionQueue(metricsOptions.metricsSubmissionUrl, applicationInfo);
        if (!uniqueEventsSubmissionQueue) {
            return undefined;
        }
        summedEventsSubmissionQueue =
            summedEventsSubmissionQueue !== null && summedEventsSubmissionQueue !== void 0 ? summedEventsSubmissionQueue : this.createSummedEventSubmissionQueue(metricsOptions.metricsSubmissionUrl, applicationInfo);
        if (!summedEventsSubmissionQueue) {
            return undefined;
        }
        return new BacktraceMetrics_1.BacktraceMetrics(metricsOptions, this._sessionProvider, this._attributeManager, summedEventsSubmissionQueue, uniqueEventsSubmissionQueue);
    }
    verifyAttributeSetup() {
        const { attributes } = this._attributeManager.get();
        const application = attributes[this.APPLICATION_ATTRIBUTE];
        const applicationVersion = attributes[this.APPLICATION_VERSION_ATTRIBUTE];
        if (!application || !applicationVersion) {
            return undefined;
        }
        return { application, applicationVersion };
    }
    createUniqueEventSubmissionQueue(metricsHost, applicationInfo) {
        var _a, _b;
        const uniqueEventsSubmissionUrl = MetricsUrlInformation_1.MetricsUrlInformation.generateUniqueEventsUrl(metricsHost, this._options.url, this._options.token);
        if (!uniqueEventsSubmissionUrl) {
            return undefined;
        }
        return new MetricsSubmissionQueue_1.MetricsSubmissionQueue(uniqueEventsSubmissionUrl, 'unique_events', this._requestHandler, {
            [this.APPLICATION_ATTRIBUTE]: applicationInfo.application,
            appversion: applicationInfo.applicationVersion,
        }, (_b = (_a = this._options) === null || _a === void 0 ? void 0 : _a.metrics) === null || _b === void 0 ? void 0 : _b.size);
    }
    createSummedEventSubmissionQueue(metricsHost, applicationInfo) {
        var _a, _b;
        const summedEventsSubmissionUrl = MetricsUrlInformation_1.MetricsUrlInformation.generateSummedEventsUrl(metricsHost, this._options.url, this._options.token);
        if (!summedEventsSubmissionUrl) {
            return undefined;
        }
        return new MetricsSubmissionQueue_1.MetricsSubmissionQueue(summedEventsSubmissionUrl, 'summed_events', this._requestHandler, {
            [this.APPLICATION_ATTRIBUTE]: applicationInfo.application,
            appversion: applicationInfo.applicationVersion,
        }, (_b = (_a = this._options) === null || _a === void 0 ? void 0 : _a.metrics) === null || _b === void 0 ? void 0 : _b.size);
    }
    optionsWithDefaults() {
        return {
            enable: true,
            autoSendInterval: this.DEFAULT_UPDATE_INTERVAL,
            metricsSubmissionUrl: 'https://events.backtrace.io',
            size: 50,
        };
    }
}
exports.MetricsBuilder = MetricsBuilder;
//# sourceMappingURL=MetricsBuilder.js.map